//
//  GUI Toolkit
//  Copyright © 2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <guit.hpp>
#include "examples.hpp"
using namespace std;
using namespace guit;


class MultiPages : public GWindow {
  GChoice channel{0}, language{1};
  GText channel_text, language_text;
  GRegText from{"00:00"}, to{"23:59"};

  GStackBox pages;
  GChoice pagenum;

public:
  MultiPages() : GWindow("#multi-pages Multi-Pages") {

    // from and to will have a HH:MM format
    from.setTimeFormat();
    to.setTimeFormat();

    auto& page0 = VBox(".page TV")
    << (ListBox()
        << "CNN" << "CNBC" << "BBC World" << "Euronews" << "Al Jazeera"
        << channel
        << channel_text
        );


    auto& page1 = VBox(".page Time & Language")
    << (Table(".panel")
        << (TRow()
            << "<b>Language"
            << (OptionBox()
                << "English" << "French" << "German" << "Cantonese"
                << language
                << language_text
                )
            )
        << (TRow() << "<b>From:" << from)
        << (TRow() << "<b>To:"   << to)
        );


    auto& page2 = VBox(".page Running")
    << (Table(".panel")
        << (TRow()
            << "<b>Channel:"
            << channel_text
            )
        << ""
        << (TRow()
            << "<b>Language:"
            << language_text
            )
        << ""
        << (TRow()
            << "<b>Time:"
            << (TextField << ("From " + ~from + " to " + ~to))
            )
        );


    pages
    << pagenum
    << page0
    << page1
    << page2;


    int last_page = int(pages.childCount()) - 1;

    auto& navbar = HBox("#navbar")
    << (~pagenum > 0)         / (Button("Previous") << [&]{pagenum--;})
    << (~pagenum < last_page) / (Button("Next")     << [&]{pagenum++;});

    *this
    << pages
    << navbar;
  }
};


GWindow& Examples::makeMultiPages() {
  return *new MultiPages;
}
