//
//  Table Example
//  Guit GUI Toolkit Examples
//  Copyright © 2019/2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <fstream>
#include <guit.hpp>
#include "examples.hpp"
using namespace std;
using namespace guit;

class MultiViewTable : public GWindow {
public:
  MultiViewTable() : GWindow("#multiview-table Multiview table")
  {
    // ** create the table and its header
    auto& table = Table("{flex:0 0}")
    << (THead()
        << ".firstname"
        << ".lastname"
        << ".gender"
        << ".age"
        << ".email"
        );

    // add rows
    for (int k = 0; k < 10; ++k) {
      table
      << (TRow()
          << Text()   // firstname
          << Text()   // lastname
          << OptionBox(".gender-options")
          << Int()    // age
          << Text()   // mail
          );
    }

    // ** create the dialog boxes
    auto& opendial = OpenDialog().setDir(":/data");
    opendial << [&]{table.read(opendial.pathname(), GIOMode::Values);};

    auto& savedial = SaveDialog().setDir("~");
    savedial << [&]{table.write(savedial.pathname(),GIOMode::Values);};

    auto& cleardial = InfoDialog("Clear","Clear Data?")
    << [&]{table.clearValues();};


    // add gadgets to the window
    *this
    << "#header Users"
    << (ScrollBox << table)
    << (ButtonBox("#controls")
        << ("  Read  " << opendial)
        << ("  Save  " << savedial)
        << ("  Print " << [&]{table.write(cout, GIOMode::Values);})
        << ("  Clear " << cleardial)
        << (" Add Row "  << [&]{addRow(table);})
        << (" New View " << [&]{addView(table);})
        );
  }

  void addRow(GTable& table) {
    auto* lastrow = table.childAt(-1);
    if (lastrow) {
      table << lastrow->clone(GClone::Children|GClone::Props);
    }
  }

  void addView(GTable& table) {
    // the window has no parent and it is not referenced by a gptr<>
    // => it will be destroyed when its close button is clicked
    auto& win = Window("New View")
    << (ScrollBox << table.clone(GClone::Sync));

    win.show();
  }
};


GWindow& Examples::makeMultiViewTable() {
  return *new MultiViewTable;
}
