//
//  Interaction Controller
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <thread>
#include <mutex>
#include <gapp.hpp>
#include <gtrigger.hpp>
#include <gvalue.hpp>
#include <ghotkey.hpp>
#include <gtimer.hpp>
#include <gtextitem.hpp>
#include <gdialog.hpp>
#include <gmenu.hpp>
#include <gcontrolimpl.hpp>
#include <ginspector.hpp>
#include <grender.hpp>
#include <gnative.hpp>
using namespace std;
namespace guit {

namespace impl {

struct ExitApp {};

struct LayoutRequest {
  LayoutRequest(GWindow* upd_win, Gadget* upd_root, bool upd_css)
  : upd_win(upd_win), upd_root(upd_root), upd_css(upd_css) {}
  gprotect<GWindow> upd_win;
  gprotect<Gadget> upd_root;
  bool upd_css;
};


struct OpenedOverlay {
  OpenedOverlay(GBox& g, Gadget* opener /*, bool autoclose, bool showaswin*/)
  : g(&g), opener(opener) {
    /*
     if (GMenu* m = g.toMenu()) {
     window = m;
     showaswin = showaswindow;
     }
     else {
     */
    window = opener ? opener->getWindow() : g.getWindow();
    softwin = true;
    this->autoclose = true;
    //}
  }
  GBox* g{};   // overlay or menu
  Gadget* opener{};
  GWindow* window{};
  bool autoclose{}, softwin{};
};


struct TipGadget {
  TipGadget();
  static void show(Gadget*);
  static void hide(Gadget*);

  void tipCB();
  GTimer timer{GPrefs::tipDelay(), 1};

  // OVERLAY MARCHE PAS => A VIRER  !!!!!
  GMenu overlay{"{background:lightgray; margin:0; border:rounded}"};
  GText text;
  gprotect<Gadget> target;
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static bool isrunning_{}, setplainmode_{}, menuisopened_{}, in_closeMenus{};
static int exitstat_{};
static GEventState keystate_{};
class GNatContext* natapp_{};
// new to avoid deleting gptrs when exiting the program
static std::list<LayoutRequest>&
layoutrequests_{*new std::list<LayoutRequest>};
static std::list<OpenedOverlay> overlays_;
static std::list<GWindow*> windows_;
static gprotect<GWindow> modalwin_;
static gptr<GWindow> mainwin_;
static GBox* menubrowsingfrom_{};
static TipGadget* tip;
static std::list<GFunction> funcalls_;
static GPoint curpos_;                // revoir par rapport aux channels !!!!
static GChannels channels_;


static void defaultError(GString const& where, GString const& msg, GObject const* obj) {
  std::cerr << "! In " << where << ": " << msg;
  // note: calling typeName() here produces: Pure virtual function called!
  if (obj) std::cerr << " (object: " << obj <<")";
  std::cerr << "\n" << std::endl;
}

static std::function<void(GString const&, GString const&, GObject const*)>
onerror_(defaultError);


static void defaultException(GString const& type, GString const& msg, bool) {
  if (GDialog::confirm(type, msg, "Continue, Abort") != 0) exit(1);
}

static std::function<void(GString const&, GString const&, bool)>
onexception_(defaultException);


TipGadget::TipGadget() {
  overlay.place_ = GMenu::MenuPlacement::BelowMouse;  // NOM IBITABLE => atMouse
  timer << [this]{tipCB();};
  overlay << (Label("{margin:0; border:none}") << text);
}

void TipGadget::show(Gadget* g) {
  //cerr << "showTip " << g << " " << g->text() << endl;
  if (!tip) tip = new TipGadget();
  tip->target = g;
  if (g) tip->timer.start();
};

void TipGadget::hide(Gadget* g) {
  //cerr << "hideTip " << g << " " << g->text() << endl;
  tip->timer.stop();
  tip->target = nullptr;
  tip->overlay.show(false);
}

void TipGadget::tipCB() {
  //if (auto* t = target->computedStyle().findProp<GTip>()) {  // !!!!
  if (auto* t = target->findProp<GTip>()) {
    overlay.show(false);
    text = t->value();
    //cerr << "showTipcb " << " tip: " << text <<endl;
    overlay.parent_ = target;
    // position when tipCB() is calleg - not when show() was called!
    overlay.show(curpos_);  // AroundEvent
  }
}

}  //endof impl

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool GControlImpl::doInspect(GWindow* win, Gadget* obj,
                             GMouseEvent& e, GPoint const& pos) {
  if (!inspector_ || win->isInternal()) return false;

  bool asked = false;

  if ((impl::keystate_ & inspecthotkey_) == inspecthotkey_) {
    asked = true;
    inspector_->show(true);
    inspector_->raise();
  }

  if (inspecting_ || asked) {
    // indirection so that that liking is possible without the GInspector module
    doinspect_(obj);
    return true;
  }
  else return false;
  /*
   Gadget* ghost{};
   if (!obj || !obj->canInspect() || !win->canInspect()
   || !(ghost = insp_->inspect(*obj))
   ) {
   return false;
   }
   // needed for moving the ghost
   pressg_[chan] = ghost;
   MouseEvent e(GTrigger::mousedown, ghost, chan, time, GEvent::NoState,
   GEvent::LeftButton, pos);
   ghost->fire(e);
   return true;
   */
}

bool GControlImpl::checkModal(GWindow* win) {
  return
  win != impl::modalwin_
  && !inspecting_
  && (impl::keystate_ & inspecthotkey_) != inspecthotkey_
  && win != inspector_;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void gsleep(GTime duration) {
  std::this_thread::sleep_for(std::chrono::milliseconds(duration));
}

void gexec(GFunction const& fun) {
  static std::mutex mutex_;
  std::lock_guard<std::mutex> lock(mutex_);
  impl::funcalls_.push_back(fun);
  impl::natapp_->pushUserEvent(&impl::funcalls_.back());
}

void gerror(GString const& context, GString const& msg, GObject const* obj) {
  if (!impl::onerror_) impl::onerror_ = impl::defaultError;
  impl::onerror_(context, msg, obj);
}

void GApp::error(GString const& context, GString const& msg, GObject const* obj) {
  gerror(context, msg, obj);
}

void GApp::onError(std::function<void(GString const&, GString const&, GObject const*)> fun) {
  impl::onerror_ = fun ? fun : impl::defaultError;
}

void GApp::onException(std::function<void(GString const&, GString const&, bool)> fun) {
  impl::onexception_ = fun ? fun : impl::defaultException;
}

void GApp::exec(GFunction const& fun) {
  gexec(fun);
}

void GApp::sleep(GTime time) {
  gsleep(time);
}

GWindow* GApp::mainWindow() {
  return impl::mainwin_;
}

GWindow* GCore::mainWindow() {
  return impl::mainwin_;
}

void GApp::close() {
  return impl::mainwin_->close();
}

void GApp::updateAll() {
  for (auto* w : impl::windows_) {
    if (w && w->isShown() && w->natwin_) w->wmodes_.mustupdate = true;
  }
}

GId GCore::getId(GString const& name, bool obtain) {
  static GId lastid{100};                  // first 100 reserverd!!!
  static std::map<GString,GId> map;

  auto it = map.find(name);
  if (it != map.end()) return it->second;
  else if (!obtain) return 0;
  else {
    lastid += 2;                            // space for ">id" !!!
    map[name] = lastid;
    return lastid;
  }
}

bool GCore::isRunning() {
  return impl::isrunning_;
}

uint32_t GCore::elapsedTime() {
  return GNatContext::elapsedTime();
}

GNatContext* GCore::natApp() {
  return impl::natapp_;
}

void GCore::subLoop(const bool& quit) {
  impl::natapp_->subLoop(quit);
}

std::list<GWindow*>& GCore::windowList() {
  return impl::windows_;
}

GWindow* GCore::modalWindow() {
  return impl::modalwin_.get();
}

void GCore::setModalWindow(GWindow* win) {
  impl::modalwin_ = win;
}

void GCore::addWindow(GWindow& win) {
  impl::windows_.push_back(&win);
}

void GCore::removeWindow(GWindow& win){
  impl::windows_.remove(&win);
  if (impl::modalwin_.get() == &win) impl::modalwin_ = nullptr;
}

bool GCore::plainMode() {
  return impl::setplainmode_;
}

void GCore::setPlainMode(bool mode) {
  impl::setplainmode_ = mode;
}

GChannels& GCore::channels() {
  return impl::channels_;
}

GChannel* GCore::channel(unsigned int id) {
  if (id < impl::channels_.size()) return impl::channels_[id];
  else return nullptr;
}

GChannel& GCore::obtainChannel() {
  return obtainChannel( (unsigned int) impl::channels_.size());
}

GChannel& GCore::obtainChannel(unsigned int id) {
  if (id >= impl::channels_.size()) impl::channels_.resize(id+1);
  if (!impl::channels_[id]) impl::channels_[id] = new GChannel(id);
  return *impl::channels_[id];
}

void GControlImpl::funCallDone() {
  impl::funcalls_.pop_front();
}

void GCore::requestUpdate() {
  //cerr << "GCore::requestUpdate"<<endl;
  // sends event so that the mainloop will update the windows
  impl::natapp_->pushUserEvent(nullptr);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GControlImpl& GControlImpl::instance() {
  static GControlImpl c;
  return c;
}

GControlImpl::GControlImpl() {
  impl::isrunning_ = false;
  impl::natapp_ = new GNatContext(*this);
}

GControlImpl::~GControlImpl() {
  //cerr << "~GControl " << mainwin_ << endl;
  delete impl::natapp_;
}


int GCore::start(GWindow* mainwin) {
  if (impl::isrunning_) {
    gerror("start","the application was already started");
    return 1;
  }
  impl::isrunning_ = true;
  impl::mainwin_ = mainwin;
  impl::mainwin_->show(true);
  
  // the event loop is restarted if there is an an exception (e.g. FatalError)
  // except if ExitApp was thrown in which vcase the app terminates
  while (true) {
    try {
      impl::natapp_->mainLoop();
    }
    catch (impl::ExitApp) {
      break;
    }
    catch (GFatalError& e) {
      impl::onexception_("Fatal Error", e.what_, true);
    }
    catch (std::runtime_error& e) {
      impl::onexception_("Runtime Error", e.what(), false);
    }
    catch (...) {
      impl::onexception_("Unknown exception", "", false);
    }
  }
  // the main window and its descendent must be destroyed NOW. one reason is
  // that some gadgets or props may be plain subobjects that will be destroyed
  // when returning from the main() function. It is safer to delete all other
  // objects now than at a later time when could point to plain subobjects
  // that would have been deleted in the meantime

  cerr << "QUIT: " << impl::mainwin_.useCount() << endl;
  impl::mainwin_ = nullptr;
  // TODO: delete all existing windows and menus that have no parent !!

  return impl::exitstat_;
}

int GApp::start(GWindow& mainwin) {
  return GCore::start(&mainwin);
}

int GApp::start(GWindow* mainwin) {
  return GCore::start(mainwin);
}

void GCore::exit(int status) {
  impl::exitstat_ = status;
  throw impl::ExitApp();
}

void GApp::exit(int status) {
  GCore::exit(status);
}

int GApp::status() {
  return impl::exitstat_;
}


namespace impl {

// can be called on an Overlay, a Menu or a MenuBar. This last case triggers
// a special mode where menus are browsed when moving the mouse
static void showOverlay(GBox& g, Gadget* opener, bool state,
                 bool autoclose, bool showaswin) {
  // ne pas modifier overlays_ si on est dans closeMenus()
  if (impl::in_closeMenus) return;

  auto it = impl::overlays_.begin();
  while (it != impl::overlays_.end()) {
    if (it->g == &g) break;
    ++it;
  }
  if (it != impl::overlays_.end()) impl::overlays_.erase(it);

  // remettre au 1er plan, donc en dernier
  if (state) impl::overlays_.
    push_back(impl::OpenedOverlay(g, opener /*,autoclose, showaswin*/));

  impl::menuisopened_ = false;
  for (auto& it2 : impl::overlays_) {
    if (it2.autoclose) {impl::menuisopened_ = true; break;}
  }
}

}

void GCore::showOverlay(GBox& g, Gadget* opener, bool state,
                        bool autoclose, bool showaswin) {
  impl::showOverlay(g, opener, state, autoclose, showaswin);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

namespace impl {

static void closeMenus(Gadget* except_opener) {
  // surtout ne pas modifier overlays_ ailleurs si on est dans closeMenus()
  impl::in_closeMenus = true;

  std::list<impl::OpenedOverlay> tmp;
  GWindow* top{};
  GBox* except_menu{};

  for (auto& it : impl::overlays_) {
    // not autoclose => keep in list of overlays
    if (!it.autoclose) tmp.push_back(it);
    else if (it.opener == except_opener) except_menu = it.g;
    else if (it.g->isShown()) {
      if (it.opener) it.opener->modes_.active = false;
      if (it.autoclose) {
        top = it.g->getWindow();
        // att cette fonction peut appeler showOverlay_, ce qui modifierait
        // overlays_, mais in_closeMenus l'empeche
        it.g->show(false);
      }
    }
  }
  impl::menuisopened_ = false;
  impl::overlays_.clear();
  impl::overlays_ = tmp;

  if (except_menu) {
    impl::menuisopened_ = true;
    impl::overlays_
    .push_back(impl::OpenedOverlay(*except_menu, except_opener /*,true, true*/));
  }

  int chan = 0;     // ???
  if (impl::channels_[chan]->browseg_) {
    impl::channels_[chan]->browseg_->modes_.active = false;
  }
  if (top) top->update();
  impl::in_closeMenus = false;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// note : box may be null
static void activateObject(GBox* box, Gadget* g) {
  if (!g || !g->isActivable()) return;
  gprotect<Gadget> boxg(box);

  if (!box) {
    if (g->isCheckable()) g->selectImpl(!g->isChecked(), Gadget::FromButton);
  }

  else if (box->modes_.hasChoice) {
    if (auto* m = box->choiceProp())
      m->choose(*box, g);
    // else missing !!!!!
  }
  else {
    if (g->isCheckable()) g->selectImpl(!g->isChecked(), Gadget::FromButton);
  }

  // fire action on gadget (already done if g->canCheck() is true)
  if (!g->isCheckable()) {
    GEvent e(on::action, GCore::elapsedTime());
    g->doFire(e);
  }

  // fire action on parent (except if choice prop bec. done)
  if (box && !box->modes_.hasChoice) {
    GEvent e(on::action, GCore::elapsedTime());
    e.aux_ = g;
    box->doFire(e);

    // propagate for table rows !!!   A REVOIR !!!
    if (box->toTRow()) {
      auto* par = box->getParent();
      // if (table && table->role() == GTable::Type().role()) table->doFire(e);
      if (par && par->toTable()) par->doFire(e);
    }
  }
}


static void clickObject(Gadget* g, GMouseEvent const& ev) {
  auto& clk = impl::channels_[ev.channel_]->clicked_;
  if (g == clk.target && ev.time_ - clk.time < GPrefs::dblclickDelay()) {
    clk.count++;
  }
  else clk.count = 1;

  clk.target = g;
  clk.time = ev.time_;

  GTrigger tr(GEventType::Click, clk.count);
  GMouseEvent e(tr, ev);
  e.clickcount_ = clk.count;
  g->doFire(e);
}


static Gadget* getActualGadget(Gadget* g, GPickMode& pick, GEvent& e,
                               GEventType mask, bool canActivate) {
  if (pick.catchedevent_) {
    e.aux_ = pick.catchedchild_;
    return g;
  }
  /*
   if (pick.through_) {
   pick.through_->doFire(e);
   }*/
  while (g) {
    if (canActivate && g->modes_.activable) return g;
    else if ((g->eventmask_ & mask) != 0) return g;
    g = g->getParent();  // parent() not parent_ !
  }
  return nullptr;
}

}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GControlImpl::mouseDown(GMouseEvent& e) {
  // cerr << "DOWN " << e.window_ << " / " << e.time_ << endl;
  if (impl::modalwin_ && checkModal(e.window_)) return;

  auto& c = obtainChannel(e.channel_);
  impl::curpos_ = e.wpos_;                    // devrait dependre du channel !

  // e.state must NOT contain the last pressed button (why not ???)
  e.state_ = c.buttonstate | impl::keystate_;
  c.buttonstate = c.buttonstate | GEventState(1 << e.button_);

  c.channelDown(e);
  if (c.buttonhandler_ && !(c.buttonhandler_)(e)) return;

  if (c.browseg_) c.browseg_->modes_.active = false;
  c.browseg_ = nullptr;

  gprotect<Gadget> curg;  // current gadget
  GPickMode pickmode;

  if (!impl::overlays_.empty()) {

    for (auto it = impl::overlays_.rbegin(); it != impl::overlays_.rend(); ++it) {
      curg = it->g->doPick(impl::curpos_, pickmode);
      if (curg) {
        if (doInspect(e.window_, curg, e, impl::curpos_)) {
          if (inspecting_ < 2 && (e.state() & GEventState::Ctrl) == 0) return;
        }
        c.pressg_ = curg;
        break;
      }
    }

    if (!curg && impl::menuisopened_) {  // not found => close all menus
      impl::closeMenus(nullptr);
      impl::menubrowsingfrom_ = {};
      return;
    }
  }

  if (!curg) {
    //cerr << "curpos_ " <<impl::curpos_ << endl;
    curg = e.window_->doPick(impl::curpos_, pickmode);
    //cerr << "item: " << curg->text() << " "  << endl;
  }

  if (curg && doInspect(e.window_, curg, e, impl::curpos_)) {
    // if ctrl is pressed the action is done normally otherwise the element
    // is just inspected (curg must not be set in this case otherwise the
    // gadget will be activated in mouseUp
    if (inspecting_ < 2 && (e.state() & GEventState::Ctrl) == 0) return;
  }

  // si l'objet ne recoit pas les MouseButton events, renvoyer cet event
  // au premier ascendant qui les traite
  c.pressg_ = impl::getActualGadget(curg, pickmode, e,
                                    GEventType::MouseButton, true);

  if (c.pressg_) {
    if (!c.pressg_->settings_.trulyEnabled) return;

    // allow browsing between menus of a menubar
    auto* par = c.pressg_->getParent();
    if (par && par->toMenuBar() && c.pressg_->modes_.activable) {
      impl::menubrowsingfrom_ = par;
    }

    if (c.pressg_->isActivable()) c.pressg_->modes_.active = true;
    e.gadget_ = c.pressg_;
    c.pressg_->update();

    if (c.pressg_->isFocusable()) {
      if (c.focusg_) c.focusg_->modes_.focus = 0;
      c.focusg_ = c.pressg_;
      c.focusg_->modes_.focus = 1;
      if (auto* tf = c.focusg_->toTextItem()) {
        tf->setCaret(e.wpos());
        tf->update();
      }
    }

    c.pressg_->doFire(e);

    if (c.pressg_->isDownActivable()) {
      impl::activateObject(c.pressg_->getParent(), c.pressg_);
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GControlImpl::mouseUp(GMouseEvent& e) {
  if (impl::modalwin_ && checkModal(e.window_)) return;

  auto& c = obtainChannel(e.channel_);
  impl::curpos_ = e.wpos_;     // devrait dependre du channel !

  // e.state contains the last pressed button
  e.state_ = c.buttonstate | impl::keystate_;
  c.buttonstate = c.buttonstate & GEventState(~(1 << e.button_));

  //cerr << "UP " << e.window_ <<" / "<< e.time_ << endl;
  c.channelUp(e);
  if (c.buttonhandler_ && !(c.buttonhandler_)(e)) return;
  if (!c.pressg_) return;

  gprotect<Gadget> curg;
  GPickMode pickmode;

  if (!impl::overlays_.empty()) {
    for (auto it = impl::overlays_.rbegin(); it != impl::overlays_.rend(); ++it) {
      curg = it->g->doPick(impl::curpos_, pickmode);

      if (c.isDragging()) {
        c.endDrag(impl::curpos_);
        return;
      }

      if (impl::menuisopened_) {
        // do action *even* if this item was not pressed (do first)
        if (curg && curg->settings_.trulyEnabled) {

          if (curg.get() == c.pressg_) {
            if ((curg->eventmask_ & GEventType::Click) != 0) impl::clickObject(curg, e);
          }

          // action on release except if the Gadget opens a menu or is text Gadget
          if (curg->isActivable()
              && !curg->settings_.hasMenu
              && !curg->isDownActivable()
              && !curg->isFocus()
              ) {
            impl::activateObject(curg->getParent(), curg);
            impl::closeMenus(nullptr);
            impl::menubrowsingfrom_ = {};
          }
          e.window_->update();
          goto END;
        }
        // none found
        if (!c.pressg_->settings_.hasMenu) {
          impl::closeMenus(nullptr);
          impl::menubrowsingfrom_ = {};
        }
      }
    }
  } // endof(!openedmenus_.empty())

  if (!curg) {     // no menu opened
    curg = e.window_->doPick(impl::curpos_, pickmode);
    if (c.isDragging()) goto END;

    curg = impl::getActualGadget(curg, pickmode, e, GEventType::MouseButton, true);

    // action only if the current widget is the pressed gadget,
    if (curg && curg.get() == c.pressg_ && curg->settings_.trulyEnabled) {

      if ((curg->eventmask_ & GEventType::Click) != 0) impl::clickObject(curg, e);

      if (!curg->settings_.hasMenu /* not a menu */
          && !curg->isDownActivable()
          /* && !curg->isFocusable() */)
        impl::activateObject(curg->getParent(), curg);
    }
  } //endof(!curg)

END:
  if (c.isDragging()) {
    c.endDrag(impl::curpos_);
    return;
  }

  if (c.pressg_) {
    e.gadget_ = c.pressg_;
    c.pressg_->doFire(e);

    if (c.pressg_->settings_.hasMenu) {
      if (c.pressg_.get() == curg.get()) impl::closeMenus(curg);
      else if (curg && curg->settings_.hasMenu) impl::closeMenus(curg);
      else {
        impl::closeMenus(nullptr);
        impl::menubrowsingfrom_ = {};
      }
    }
    if (c.pressg_->isActive() && !impl::menuisopened_) {
      c.pressg_->modes_.active = false;
      c.pressg_->update();
    }
  }

  if (curg && curg->isActive() && (!curg->settings_.hasMenu || !impl::menuisopened_)) {
    curg->modes_.active = false;
    curg->update();
  }

  c.pressg_ = nullptr;
  if (c.browseg_) c.browseg_->modes_.active = false;
  c.browseg_ = nullptr;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GControlImpl::mouseMove(GMouseEvent& e) {
  //cerr << "\nMOVE win " << e.window_ << endl;
  if (impl::modalwin_ && checkModal(e.window_)) return;   // interdit les tooltips !

  auto& c = obtainChannel(e.channel_);
  e.state_ = c.buttonstate | impl::keystate_;
  impl::curpos_ = e.wpos_;             // devrait dependre de channel !!!

  if (c.maxderive_ > 0) {
    c.derive_ = elapsedTime() - e.time_ - c.derive_;
    // cerr << "derive for #" << c.id_
    // << " = " << c.derive_ << " / delta " << (elapsedTime() - e.time_) << endl;
    if (c.derive_ <= 0) {
      c.derive_ = 0;
      c.derivecount_ = 0;
    }
    else if (c.derive_ > c.maxderive_) {
      if (c.derivecount_++ > 5) {
        c.derivecount_ = 0;
        return;
      }
    }
  }
  c.channelMove(e);
  if (c.movehandler_ && !(c.movehandler_)(e)) return;

  gprotect<Gadget> curg;
  GPickMode pickmode;

  // send mousedrag to last pressed object
  if (c.pressg_ && (c.pressg_->eventmask_ & GEventType::MouseDrag) != 0) {
    GMouseEvent e2(GTrigger::mousedrag,e);
    c.pressg_->doFire(e2);
  }

  if (!impl::overlays_.empty()) {  // a menu is opened
    for (auto it = impl::overlays_.rbegin(); it != impl::overlays_.rend(); ++it) {
      curg = it->g->doPick(impl::curpos_, pickmode);
      if (c.isDragging()) {
        //c.move(curpos_ + e.window_->pos());
        return;
      }
      if (curg && impl::menuisopened_) {
        if (c.browseg_) c.browseg_->modes_.active = false;
        c.browseg_ = nullptr;
        auto* par = curg->getParent();
        if (curg->isActivable() && par && !par->toMenuBar()) {
          c.browseg_ = curg;
          c.browseg_->modes_.active = true;
        }
        e.window_->update();
        break;
      }
    }
  }

  if (!curg) {
    curg = e.window_->doPick(impl::curpos_, pickmode);
  }

  if (c.isDragging()) {
    //c.move(curpos_ + e.window_->pos());
    return;
  }

  if (curg) {

    if ((curg->eventmask_ & GEventType::MouseMove) != 0) curg->doFire(e);

    if (c.enterg_ && c.enterg_.get() != curg.get()) {
      //static int n; cerr << "leave "<< ++n << " " << c.enterg_  <<endl;
      if ((c.enterg_->eventmask_ & GEventType::MouseCross) != 0) {
        GMouseEvent e2(GTrigger::mouseout, e);
        c.enterg_->doFire(e2);
      }
      if (c.enterg_->modes_.hasTip /* && c.ident() == 0*/) {
        impl::TipGadget::hide(c.enterg_);
      }
      if (impl::menuisopened_ && curg->settings_.hasMenu) {
        // browse between menus of a menubar (but dont browse other menus)
        if (impl::menubrowsingfrom_
            && curg->getParent() == impl::menubrowsingfrom_) {
          impl::closeMenus(nullptr);
          curg->modes_.active = true;     // needed after closeOtherMenus()
          GMouseEvent e2(GTrigger::mousedown, e);
          curg->doFire(e2);   // opens the menu associated to this item
          e.window_->update();
        }
      }

      c.enterg_ = nullptr;
    }

    if (!c.enterg_) {
      c.enterg_ = curg;
      if ((c.enterg_->eventmask_ & GEventType::MouseCross) != 0){
        GMouseEvent e2(GTrigger::mouseover, e);
        c.enterg_->doFire(e2);
      }
      if (c.enterg_->modes_.hasTip /* && c.ident() == 0 */) {
        impl::TipGadget::show(c.enterg_);
      }
    }

    /*
     if (impl::menuisopened_ && curg->settings_.hasMenu) {
     closeOtherMenus(nullptr);          // close all menus except menubar
     curg->modes_.activated = true;     // needed after closeOtherMenus()
     GMouseEvent e2(GTrigger::mousedown, e);
     curg->fire(e2);
     e.window_->update();
     }
     else */

    if (c.pressg_
        && c.pressg_.get() == curg.get()
        && c.pressg_.get() == c.focusg_.get())
    {
      if (auto* tf = c.focusg_->toTextItem()) {
        GMouseEvent e2(GTrigger::mousedrag, e);
        e2.gadget_ = c.pressg_;   // pressg not curg!

        //cerr <<  "PRESS " <<c.pressg_ << endl;
        tf->moveCaret(e2.wpos());
        tf->update();
      }
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GControlImpl::wheelMove(GMouseEvent& e) {
  auto& c = obtainChannel(0);
  if (c.wheelhandler_ && !(c.wheelhandler_)(e)) return;

  GPickMode pickmode;
  Gadget* g = nullptr;

  if (!impl::overlays_.empty()) {
    for (auto it = impl::overlays_.rbegin(); it != impl::overlays_.rend(); ++it) {
      g = it->g->doPick(impl::curpos_, pickmode);
      if (g) break;
    }
  }
  else {
    g = e.window_->doPick(impl::curpos_, pickmode);
  }

  if ((g = impl::getActualGadget(g, pickmode, e, GEventType::Wheel, false))) {
    if (auto* sb = g->toScrollBox()) {
      sb->setScrollDelta(e.dx_ * GPrefs::wheelGain(),
                         e.dy_ * GPrefs::wheelGain());
    }
    g->doFire(e);
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GControlImpl::keyDown(GKeyEvent& e) {
  if (impl::modalwin_ && checkModal(e.window_)) return;
  auto& c = obtainChannel(0);

  // e.state does not contain the last pressed key
  e.state_ = impl::keystate_;
  impl::keystate_ = impl::keystate_ | GEvent::modifierToState(e.key_);

  if (c.keyhandler_ && !(c.keyhandler_)(e)) return;

  if (c.focusg_ && c.focusg_->settings_.trulyEnabled) {
    c.focusg_->doFire(e);
    if (auto* tf = c.focusg_->toTextItem()) {
      if (tf->keyCB(e)) {
        tf->update();
        return;
      }
    }
  }

  if (e.window_->hotkeys_) {
    if (e.window_->triggerHotkey(impl::keystate_, e.key_)) return;
  }

  // otherwise, if hotkey not found, send event to 1st gadget that can receive it
  // autre solution;: envoyer a window?
  gprotect<Gadget> g;
  GPickMode pickmode;
  g = e.window_->doPick(impl::curpos_, pickmode);
  g = impl::getActualGadget(g, pickmode, e, GEventType::Key, false);
  if (g) g->doFire(e);
}


void GControlImpl::keyUp(GKeyEvent& e) {
  if (impl::modalwin_ && checkModal(e.window_)) return;
  auto& c = obtainChannel(0);

  // e.state contains the last pressed key
  e.state_ = impl::keystate_;
  //keystate_ &= ~ GEvent::modifierToState(e.key_);
  impl::keystate_ = impl::keystate_
  & GEventState( ~(uint32_t) GEvent::modifierToState(e.key_) );

  if (c.keyhandler_ && !(c.keyhandler_)(e)) return;

  if (c.focusg_ && c.focusg_->settings_.trulyEnabled) {
    c.focusg_->doFire(e);
  }
  /*
  else {
    gprotect<Gadget> g;
    GPickMode pickmode;
    g = e.window_->pick(curpos_, pickmode);
    g = getActualGadget(g, pickmode, e, GTrigger::KeyMask, false);
    if (g) g->fire(e);
  }*/
}


void GControlImpl::keyType(GKeyEvent& e) {
  if (impl::modalwin_ && checkModal(e.window_)) return;
  auto& c = obtainChannel(0);
  e.state_ = impl::keystate_;

  if (c.keyhandler_ && !(c.keyhandler_)(e)) return;

  if (c.focusg_ && c.focusg_->settings_.trulyEnabled) {
    c.focusg_->doFire(e);

    if (auto* tf = c.focusg_->toTextItem()) {
      if (tf->keyCB(e)) tf->update();
    }
  }
}


void GControlImpl::touchEvent(GTouchEvent& e) {
  //cerr << "FINGER event / type " << type << endl;
  if ((impl::mainwin_->eventmask_ & GEventType::Touch) != 0) {
    impl::mainwin_->doFire(e);
  }
}

void GControlImpl::gestureEvent(GGestureEvent& e) {
  //cerr << "GESTURE event / type " << type << endl;
  if ((impl::mainwin_->eventmask_ & GEventType::Gesture) != 0) {
    impl::mainwin_->doFire(e);
  }
}

void GControlImpl::windowResized(GWindow& win, GTime t, float w, float h) {
  win.resize(w, h);     // on peut tjrs retailler interactivement
  GEvent e(GTrigger::resize, t);
  win.doFire(e);
}

void GControlImpl::windowMoved(GWindow& win,  GTime t, float x, float y) {
  win.prop<GPos>().set(x,y);
  /* normalement c'est fait par GWindow::moveImpl appele par CB    $$$$$
  GEvent e(GTrigger::move, t);
  win.doFire(e);
   */
}

void GControlImpl::windowShown(GWindow& win, GTime t) {
  //cerr << "windowShown " << win << endl;
}

void GControlImpl::windowHidden(GWindow& win, GTime t) {
  // modal windows must not be hidden
  if (impl::modalwin_ && impl::modalwin_.get() == &win) win.restore();
}

void GControlImpl::windowMaximized(GWindow& win, GTime t) {
  // modal windows must not be maximized
  if (impl::modalwin_.get() == &win) win.restore();
}

void GControlImpl::windowMinimized(GWindow& win, GTime t) {
  // modal windows must not be minimized
  if (impl::modalwin_.get() == &win) win.restore();
}

void GControlImpl::windowClosed(GWindow& win, GTime t) {
  if (impl::modalwin_.get() == &win) impl::modalwin_ = nullptr;
  win.close();
}

void GControlImpl::windowEntered(GWindow& win, GTime t) {
  GMouseEvent e(on::mouseover, t);
  e.window_ = &win;
  obtainChannel(0).enterg_ = &win;
  win.doFire(e);
}

void GControlImpl::windowExited(GWindow& win, GTime t) {
  GMouseEvent e(on::mouseout, t);
  e.window_ = &win;
  win.doFire(e);
  obtainChannel(0).enterg_ = nullptr;
}

void GControlImpl::windowfocusIn(GWindow& win, GTime t) {
  GMouseEvent e(on::focus, t);
  e.window_ = &win;
  //obtainChannel(0).focusg_ = &win;
  win.doFire(e);

  // modal windows must be always on top of other windows
  if (impl::modalwin_ && checkModal(&win)) impl::modalwin_->raise();
  //obtainChannel(0).focusg_ = &win;
}

void GControlImpl::windowfocusOut(GWindow& win, GTime t) {
  //cerr << "FOCUS-OUT " << win << endl;
  GMouseEvent e(on::focusout, t);
  win.doFire(e);
  if (win.wmodes_.autoclose) win.show(false);
  //obtainChannel(0).focusg_ = nullptr;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

namespace impl {

static void updateLayout(GWindow* upd_win, Gadget* upd_root, bool upd_css) {
  // already asked
  //if (upd_win->wmodes_.mustlayout) return;

  // already asked sauf cas particulier des menus sinon ils seront jamais
  // updates car ils sont pas directemnt dans l'arbre
  if (upd_win->wmodes_.mustlayout && (!upd_root || !upd_root->toMenu()))
    return;

  if (!upd_root || upd_win == upd_root) {  // must layout entire window
    upd_win->wmodes_.mustlayout = true;

    // remove previous partial requests if any
    auto it = impl::layoutrequests_.begin();
    while (it != impl::layoutrequests_.end()) {
      if (it->upd_win.get() == upd_win) {
        auto itdel = it++;
        impl::layoutrequests_.erase(itdel);
      }
      else ++it;
    }
  }
  // verifier qu'il y a pas deja le meme et ceux contenus les uns dans les autres !!                 !!!
  impl::layoutrequests_.push_back(impl::LayoutRequest(upd_win, upd_root, upd_css));
  upd_win->update();
}
}

void GCore::updateLayout(GWindow* upd_win, Gadget* upd_root, bool upd_css) {
  impl::updateLayout(upd_win, upd_root, upd_css);
}

void GControlImpl::paintWindow(GWindow* win) {
  //cerr << "GControl::paintWindow " << win << endl;
  // paint() must not call itself. This problem may happen when NatWin::setSize()
  // is indirectly called by paint():
  // NatWin::setSize() calls SDL_SetWindowSize(), which may trigger Cocoa
  // windowDidResize, which calls natwin::paintCB(), which calls this function
  static bool in_paint{};
  if (in_paint) return;
  in_paint = true;

  // update the layout first, if needed
  for (auto& it : impl::layoutrequests_) {
    // must layout entire window
    if (!it.upd_root || it.upd_win.get() == it.upd_root.get()) {
      if (it.upd_css) it.upd_win->topInit();
      it.upd_win->updateLayout(Gadget::Update::Now);
      it.upd_win->wmodes_.mustlayout = false;
    }
    else {   // partial layout update
      it.upd_root->updateLayout(Gadget::Update::Now);
    }
  }

  impl::layoutrequests_.clear();
  GRenderGraphics dc(*win);

  win->doPaint(dc, nullptr);
  for (auto& it : impl::overlays_) {
    // it.opened can be an overlay, a menu or a menubar
    if (it.softwin && it.window == win && it.g->isShown()) {
      GRenderGraphics dc2(*win);
      it.g->doPaint(dc2, nullptr);
    }
  }
  if (channels().size() > 1) GCursor::drawCursors(dc);
  win->wmodes_.mustupdate = false;
  in_paint = false;
}


void GCore::paintAll() {
  //cerr << "GControl::paintAll " << endl;
  for (auto* w : impl::windows_) {
    if (w->wmodes_.mustupdate && w->natwin_) {
      //cerr << " - paint win: "<< w << endl;
      w->natwin_->paint();
    }
  }
}

void GApp::paintAll() {
  GCore::paintAll();
}


void GCore::printClassSizes() {
  cout
  << "Object: " << sizeof(GObject) << "\n"
  << "Prop: " << sizeof(GProp)  << "\n"
  << "PropList: " << sizeof(GPropList) << "\n"
  << "VarProp: " << sizeof(GVarProp)  << "\n"
  << "Text: " << sizeof(GText)  << "\n"
  << "Color: " << sizeof(GColor)  << "\n"
  << "Font: " << sizeof(GFont) << "\n"
  << "---" <<endl;
  cout
  << "Gadget: " << sizeof(Gadget) << "\n"
  << "GadgetModes: " << sizeof(GadgetModes) << "\n"
  << "GadgetLayout: " << sizeof(GadgetLayout) << "\n"
  << "GEventType: " << sizeof(GEventType) << "\n"
  << "Rect " << sizeof(GRect) << "\n"
  << "Item: " << sizeof(GItem)<< "\n"
  << "TextItem: " << sizeof(GTextItem)<< "\n"
  << "Box: " << sizeof(GBox) << "\n"
  << "---" <<endl;
  cout
  << "ssize_t: " << sizeof(ssize_t) << "\n"
  << "std::string: " << sizeof(std::string) << "\n"
  << "std::vector: " << sizeof(std::vector<void*>) << "\n"
  << "std::list: " << sizeof(std::list<void*>) << "\n"
  << "std::forward_list: " << sizeof(std::forward_list<void*>) << "\n"
  << "std::map: " << sizeof(std::map<std::string,void*>) << "\n"
  <<endl;
}

}
