//
//  Viewport for scrolling another gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_ScrollBox_hpp
#define Guit_ScrollBox_hpp
#include <gbox.hpp>
#include <gslider.hpp>
namespace guit {

class GScrollBar;

/// Viewport for displaying a subpart of another gadget.
/// A ScrollBox should have only one child: the gadget that it will (partly) display. Scrollbars are
/// automatically created to allow choosing which part of the child is shown.
///
/// Important: a ScrollBox must have a size (see GSize) and/or an appropriate flex value (see GFlex),
/// because it can't deduce its size form the child it contains.
/// Moreover, the behavior of the child when resizing the ScrollBox depends on the child 's flex value.
/// For instance, if this value is "0 0" the child size won't change, if it's "1 1" it will increase (in both
/// directions), but won't decrease. see GFlex.
class GScrollBox : public GBox {
protected:
  GScrollBox(GadgetType&, GString const& css_and_title);

public:
  ~GScrollBox();
  GScrollBox(GString const& css_and_title = "");
  GScrollBox& clone(GClone const&) override;

  GScrollBox* toScrollBox() override {return this;}

  /// Scrolls the scrollbox (values should be between 0 and 1).
  void setScroll(float hscroll, float vscroll);
  
  /// Scrolls the scrollbox by thsi amount.
  void setScrollDelta(float dx, float dy);
  
  /// Returns horizontal and vertical scrolling values.
  /// @{
  float hscroll() const;
  float vscroll() const;
  /// @}

  /// returns the horizontal and vertical scrollbars.
  /// @{
  GScrollBar& hscrollbar();
  GScrollBar& vscrollbar();
  /// @}

  using GType = GadgetType_<GScrollBox,GBox>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  void pack(GRender* parent_render) override;
  GRect scrollImpl(GRender& r, float dx, float dy);
  gptr<GScrollBar> hscrollbar_, vscrollbar_;
  /// @}
};

/// Creates a ScrollBox.
GScrollBox& ScrollBox(GString const& css_and_title = "");

}
#endif
