//
//  Boxes that stack gadgets
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_StackBox_hpp
#define Guit_StackBox_hpp
#include <gchoicebox.hpp>
namespace guit {

/// Box that displays only one of its children at a time.
/// Works as GChoiceBox, the choose() method selects the child that is displayed, etc.
/// @see GChoiceBox (parent class), GTabBox (child class).
class GStackBox : public GChoiceBox {
protected:
  GStackBox(GadgetType&, GString const& css_and_title);

public:
  GStackBox(GString const& css_and_title = "");
  GStackBox& clone(GClone const&) override;

  using GType = GadgetType_<GStackBox,GChoiceBox>;
  static GType& Type();
  GType& type() const override {return Type();}
};

GStackBox& StackBox(GString const& css_and_title = "");



/// Box that displays only one of its children at a time and a tab bar that controls which child is displayed.
/// The title of each tab is the title (or GTitle prop) of the corresponding child.
/// Works as GChoiceBox, the choose() method selects the child that is displayed, etc.
/// @see GStackBox.
class GTabBox : public GStackBox {
protected:
  GTabBox(GadgetType&, GString const& css_and_title);

public:
  ~GTabBox();
  GTabBox(GString const& css_and_title = "");
  GTabBox& clone(GClone const&) override;

  GChoiceBox& tabs();

  using GChoiceBox::add;
  void add(Gadget* child, int index = -1) override;
  void add(Gadget& child, int index = -1) override {GTabBox::add(&child, index);}

  using GType = GadgetType_<GTabBox,GStackBox>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  void setChoiceImpl(GChoice& c, bool add_to_list) override;
  gptr<GChoiceBox> tabs_;
};

GTabBox& TabBox(GString const& css_and_title = "");

}
#endif
